"""Helpers for setting a cooldown on commands."""

from __future__ import annotations

import asyncio
import random
import time
import typing
import weakref
from collections.abc import Awaitable, Callable, Hashable, Iterable
from contextlib import suppress
from dataclasses import dataclass

import discord
from discord.ext.commands import CommandError, Context

from pydis_core.utils import scheduling
from pydis_core.utils.function import command_wraps

__all__ = ["CommandOnCooldown", "P", "R", "block_duplicate_invocations"]

_KEYWORD_SEP_SENTINEL = object()

_ArgsList = list[object]
_HashableArgsTuple = tuple[Hashable, ...]

if typing.TYPE_CHECKING:
    from pydis_core import BotBase

P = typing.ParamSpec("P")
"""The command's signature."""
R = typing.TypeVar("R")
"""The command's return value."""


class CommandOnCooldown(CommandError, typing.Generic[P, R]):  # noqa: N818
    """Raised when a command is invoked while on cooldown."""

    def __init__(
        self,
        message: str | None,
        function: Callable[P, Awaitable[R]],
        /,
        *args: P.args,
        **kwargs: P.kwargs,
    ):
        super().__init__(message, function, args, kwargs)
        self._function = function
        self._args = args
        self._kwargs = kwargs

    async def call_without_cooldown(self) -> R:
        """
        Run the command this cooldown blocked.

        Returns:
            The command's return value.
        """
        return await self._function(*self._args, **self._kwargs)


@dataclass
class _CooldownItem:
    non_hashable_arguments: _ArgsList
    timeout_timestamp: float


@dataclass
class _SeparatedArguments:
    """Arguments separated into their hashable and non-hashable parts."""

    hashable: _HashableArgsTuple
    non_hashable: _ArgsList

    @classmethod
    def from_full_arguments(cls, call_arguments: Iterable[object]) -> typing.Self:
        """Create a new instance from full call arguments."""
        hashable = list[Hashable]()
        non_hashable = list[object]()

        for item in call_arguments:
            try:
                hash(item)
            except TypeError:
                non_hashable.append(item)
            else:
                hashable.append(item)

        return cls(tuple(hashable), non_hashable)


class _CommandCooldownManager:
    """
    Manage invocation cooldowns for a command through the arguments the command is called with.

    Use `set_cooldown` to set a cooldown,
    and `is_on_cooldown` to check for a cooldown for a channel with the given arguments.
    A cooldown lasts for `cooldown_duration` seconds.
    """

    def __init__(self, *, cooldown_duration: float):
        self._cooldowns = dict[tuple[Hashable, _HashableArgsTuple], list[_CooldownItem]]()
        self._cooldown_duration = cooldown_duration
        self.cleanup_task = scheduling.create_task(
            self._periodical_cleanup(random.uniform(0, 10)),
            name="CooldownManager cleanup",
        )
        weakref.finalize(self, self.cleanup_task.cancel)

    def set_cooldown(self, channel: Hashable, call_arguments: Iterable[object]) -> None:
        """Set `call_arguments` arguments on cooldown in `channel`."""
        timeout_timestamp = time.monotonic() + self._cooldown_duration
        separated_arguments = _SeparatedArguments.from_full_arguments(call_arguments)
        cooldowns_list = self._cooldowns.setdefault(
            (channel, separated_arguments.hashable),
            [],
        )

        for item in cooldowns_list:
            if item.non_hashable_arguments == separated_arguments.non_hashable:
                item.timeout_timestamp = timeout_timestamp
                return

        cooldowns_list.append(_CooldownItem(separated_arguments.non_hashable, timeout_timestamp))

    def is_on_cooldown(self, channel: Hashable, call_arguments: Iterable[object]) -> bool:
        """Check whether `call_arguments` is on cooldown in `channel`."""
        current_time = time.monotonic()
        separated_arguments = _SeparatedArguments.from_full_arguments(call_arguments)
        cooldowns_list = self._cooldowns.get(
            (channel, separated_arguments.hashable),
            [],
        )

        for item in cooldowns_list:
            if item.non_hashable_arguments == separated_arguments.non_hashable:
                return item.timeout_timestamp > current_time
        return False

    async def _periodical_cleanup(self, initial_delay: float) -> None:
        """
        Delete stale items every hour after waiting for `initial_delay`.

        The `initial_delay` ensures cleanups are not running for every command at the same time.
        A strong reference to self is only kept while cleanup is running.
        """
        weak_self = weakref.ref(self)
        del self

        await asyncio.sleep(initial_delay)
        while True:
            await asyncio.sleep(60 * 60)
            weak_self()._delete_stale_items()  # noqa: SLF001

    def _delete_stale_items(self) -> None:
        """Remove expired items from internal collections."""
        current_time = time.monotonic()

        for key, cooldowns_list in self._cooldowns.copy().items():
            filtered_cooldowns = [
                cooldown_item for cooldown_item in cooldowns_list if cooldown_item.timeout_timestamp < current_time
            ]

            if not filtered_cooldowns:
                del self._cooldowns[key]
            else:
                self._cooldowns[key] = filtered_cooldowns


def _create_argument_tuple(*args: object, **kwargs: object) -> tuple[object, ...]:
    return (*args, _KEYWORD_SEP_SENTINEL, *kwargs.items())


def block_duplicate_invocations(
    *,
    cooldown_duration: float = 5,
    send_notice: bool = False,
    args_preprocessor: Callable[P, Iterable[object]] | None = None,
) -> Callable[[Callable[P, Awaitable[R]]], Callable[P, Awaitable[R]]]:
    """
    Prevent duplicate invocations of a command with the same arguments in a channel for ``cooldown_duration`` seconds.

    Args:
        cooldown_duration: Length of the cooldown in seconds.
        send_notice: If :obj:`True`, notify the user about the cooldown with a reply.
        args_preprocessor: If specified, this function is called with the args and kwargs the function is called with,
                            its return value is then used to check for the cooldown instead of the raw arguments.

    Returns:
        A decorator that adds a wrapper which applies the cooldowns.

    Warning:
        The created wrapper raises :exc:`CommandOnCooldown` when the command is on cooldown.
    """

    def decorator(func: Callable[P, Awaitable[R]]) -> Callable[P, Awaitable[R]]:
        mgr = _CommandCooldownManager(cooldown_duration=cooldown_duration)

        @command_wraps(func)
        async def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
            if args_preprocessor is not None:
                all_args = args_preprocessor(*args, **kwargs)
            else:
                all_args = _create_argument_tuple(*args[2:], **kwargs)  # skip self and ctx from the command
            ctx = typing.cast("Context[BotBase]", args[1])

            if not isinstance(ctx.channel, discord.DMChannel):
                if mgr.is_on_cooldown(ctx.channel, all_args):
                    if send_notice:
                        with suppress(discord.NotFound):
                            await ctx.reply("The command is on cooldown with the given arguments.")
                    raise CommandOnCooldown(ctx.message.content, func, *args, **kwargs)
                mgr.set_cooldown(ctx.channel, all_args)

            return await func(*args, **kwargs)

        return wrapper

    return decorator



import json

from aiohttp import ClientConnectorError, ClientSession
from aiohttp.web import HTTPException
from pydantic import BaseModel

from pydis_core.utils import logging

log = logging.get_logger(__name__)

DEFAULT_PASTEBIN = "https://paste.pythondiscord.com"
FAILED_REQUEST_ATTEMPTS = 3
MAX_PASTE_SIZE = 512 * 1024  # 512kB
"""The maximum allows size of a paste, in bytes."""

# A dict where the keys are paste services and the keys are the lists of lexers that paste service supports.
_lexers_supported_by_pastebin: dict[str, list[str]] = {}


class PasteResponse(BaseModel):
    """
    A successful response from the paste service.

    Args:
        link (:obj:`str`): The URL to the saved paste.
        removal (:obj:`str`): The URL to delete the saved paste.
    """

    link: str
    removal: str


class PasteFile(BaseModel):
    """
    A file to be pasted to the paste service.

    Args:
        content (:obj:`str`): The content of the file.
        name (:obj:`str`): The name of the file to upload. Defaults to ``""``
        lexer (:obj:`str`): The lexer to use when applying text formatting. Defaults to ``"python"``
    """

    content: str
    name: str = ""
    lexer: str = "python"


class PasteUploadError(Exception):
    """Raised when an error is encountered uploading to the paste service."""


class PasteUnsupportedLexerError(Exception):
    """Raised when an unsupported lexer is used."""


class PasteTooLongError(Exception):
    """Raised when content is too large to upload to the paste service."""


def _validate_file(file: PasteFile, paste_url: str, max_size: int) -> None:
    if file.lexer not in _lexers_supported_by_pastebin[paste_url]:
        raise PasteUnsupportedLexerError(f"Lexer '{file.lexer}' not supported by pastebin.")

    content_size = len(file.content.encode())
    if content_size > max_size:
        log.info("Contents too large to send to paste service.")
        raise PasteTooLongError(f"Contents of size {content_size} greater than maximum size {max_size}")


async def send_to_paste_service(
    *,
    files: list[PasteFile],
    http_session: ClientSession,
    paste_url: str = DEFAULT_PASTEBIN,
    max_size: int = MAX_PASTE_SIZE,
) -> PasteResponse:
    """
    Upload some contents to the paste service.

    Args:
        files: The files to be uploaded to the paste service.
        http_session (aiohttp.ClientSession): The session to use when POSTing the content to the paste service.
        paste_url: The base url to the paste service.
        max_size: The max number of bytes to be allowed. Anything larger than :obj:`MAX_PASTE_SIZE` will be rejected.

    Raises:
        :exc:`ValueError`: ``max_length`` greater than the maximum allowed by the paste service.
        :exc:`PasteTooLongError`: ``contents`` too long to upload.
        :exc:`PasteUploadError`: Uploading failed.

    Returns:
        A pydantic model containing both the URL of the paste, and a URL to remove the paste.
    """
    if max_size > MAX_PASTE_SIZE:
        raise ValueError(f"`max_length` must not be greater than {MAX_PASTE_SIZE}")

    if paste_url not in _lexers_supported_by_pastebin:
        try:
            async with http_session.get(f"{paste_url}/api/v1/lexer") as response:
                response_json = await response.json()  # Supported lexers are the keys.
        except HTTPException:
            raise PasteUploadError("Could not fetch supported lexers from selected paste_url.")

        _lexers_supported_by_pastebin[paste_url] = list(response_json)

    for file in files:
        _validate_file(file, paste_url, max_size)

    payload = {
        "expiry": "30days",
        "long": "on",  # Use a longer URI for the paste.
        "files": [file.model_dump() for file in files]
    }
    for attempt in range(1, FAILED_REQUEST_ATTEMPTS + 1):
        try:
            async with http_session.post(f"{paste_url}/api/v1/paste", json=payload) as response:
                response_text = await response.text()
        except ClientConnectorError:
            log.warning(
                f"Failed to connect to paste service at url {paste_url}, "
                f"trying again ({attempt}/{FAILED_REQUEST_ATTEMPTS})."
            )
            continue
        except Exception:
            log.exception(
                f"An unexpected error has occurred during handling of the request, "
                f"trying again ({attempt}/{FAILED_REQUEST_ATTEMPTS})."
            )
            continue

        if "Text exceeds size limit" in response_text:
            log.info("Contents too large to send to paste service after lexing.")
            raise PasteTooLongError(
                f"Contents of file greater than maximum size {max_size} after lexing."
            )

        response_json = json.loads(response_text)
        if response.status == 400:
            log.warning(
                f"Paste service returned error {response_json['message']} with status code {response.status}, "
                f"trying again ({attempt}/{FAILED_REQUEST_ATTEMPTS})."
            )
            continue

        if response.status == 200:
            log.info(f"Successfully uploaded contents to {response_json['link']}.")
            return PasteResponse(link=response_json["link"], removal=response_json["removal"])

        log.warning(
            f"Got unexpected JSON response from paste service: {response_json}\n"
            f"trying again ({attempt}/{FAILED_REQUEST_ATTEMPTS})."
        )

    raise PasteUploadError("Failed to upload contents to paste service")



"""Useful utilities and tools for Discord bot development."""

import pydis_core.utils.error_handling.commands as command_error_handling_module
import pydis_core.utils.error_handling.commands.abc as error_handling_abstractions
import pydis_core.utils.error_handling.commands.manager as command_error_manager
from pydis_core.utils import (
    _monkey_patches,
    caching,
    channel,
    checks,
    commands,
    cooldown,
    error_handling,
    function,
    interactions,
    lock,
    logging,
    members,
    messages,
    pagination,
    paste_service,
    regex,
    scheduling,
)
from pydis_core.utils._extensions import unqualify


def apply_monkey_patches() -> None:
    """
    Applies all common monkey patches for our bots.

    Patches :obj:`discord.ext.commands.Command` and :obj:`discord.ext.commands.Group` to support root aliases.
        A ``root_aliases`` keyword argument is added to these two objects, which is a sequence of alias names
        that will act as top-level groups rather than being aliases of the command's group.

        It's stored as an attribute also named ``root_aliases``

    Patches discord's internal ``send_typing`` method so that it ignores 403 errors from Discord.
        When under heavy load Discord has added a CloudFlare worker to this route, which causes 403 errors to be thrown.
    """
    _monkey_patches._apply_monkey_patches()  # noqa: SLF001


__all__ = [
    apply_monkey_patches,
    caching,
    channel,
    checks,
    commands,
    command_error_handling_module,
    error_handling_abstractions,
    command_error_manager,
    cooldown,
    error_handling,
    function,
    interactions,
    lock,
    logging,
    members,
    messages,
    pagination,
    paste_service,
    regex,
    scheduling,
    unqualify,
]

__all__ = [module.__name__ for module in __all__]



from collections.abc import Callable
from typing import NoReturn

from discord import Interaction
from discord.ext.commands import Context

from pydis_core.utils.error_handling.commands import AbstractCommandErrorHandler
from pydis_core.utils.logging import get_logger

log = get_logger(__name__)


class CommandErrorManager:
    """A class that registers error handlers and handles all command related errors."""

    def __init__(self, default: AbstractCommandErrorHandler):
        self._handlers = []
        self._registered_handlers = set()
        self._default = default

    async def handle_error(
        self,
        error: Exception,
        context_or_interaction: Context | Interaction
    ) -> None:
        """
        Handle a Discord exception.

        Iterate through available handlers by registration order, and choose the first one capable of handling
        the error as determined by `should_handle_error`; there is no priority system.
        """
        error = getattr(error, "original", error)
        for handler in self._handlers + [self._default]:
            if await handler.should_handle_error(error):
                callback = self._get_callback(handler, context_or_interaction)
                await callback(context_or_interaction, error)
                break

    def register_handler(self, handler: AbstractCommandErrorHandler) -> None:
        """Register a command error handler."""
        handler_name = type(handler).__name__
        if handler_name in self._registered_handlers:
            log.info(f"Skipping registration of command error handler {handler_name!r} as it's already registered.")
            return

        self._handlers.append(handler)
        self._registered_handlers.add(handler_name)

    @staticmethod
    def _get_callback(
        handler: AbstractCommandErrorHandler,
        context_or_interaction: Context | Interaction
    ) -> Callable[[Context, Exception], NoReturn] | Callable[[Interaction, Exception], NoReturn] | None:
        """Get the handling callback that will be used."""
        if isinstance(context_or_interaction, Context):
            return handler.handle_text_command_error
        if isinstance(context_or_interaction, Interaction):
            return handler.handle_app_command_error
        raise ValueError(f"Expected Context or Interaction, got {type(context_or_interaction).__name__}")



"""An async transport method for statsd communication."""

import asyncio
import socket

from statsd.client.base import StatsClientBase

from pydis_core.utils import scheduling


class AsyncStatsClient(StatsClientBase):
    """An async implementation of :obj:`statsd.client.base.StatsClientBase` that supports async stat communication."""

    def __init__(
        self,
        loop: asyncio.AbstractEventLoop,
        host: str = "localhost",
        port: int = 8125,
        prefix: str | None = None
    ):
        """
        Create a new :obj:`AsyncStatsClient`.

        Args:
            loop (asyncio.AbstractEventLoop): The event loop to use when creating the
                :obj:`asyncio.loop.create_datagram_endpoint`.
            host: The host to connect to.
            port: The port to connect to.
            prefix: The prefix to use for all stats.
        """
        _, _, _, _, addr = socket.getaddrinfo(
            host, port, socket.AF_INET, socket.SOCK_DGRAM
        )[0]
        self._addr = addr
        self._prefix = prefix
        self._loop = loop
        self._transport: asyncio.DatagramTransport | None = None

    async def create_socket(self) -> None:
        """Use :obj:`asyncio.loop.create_datagram_endpoint` from the loop given on init to create a socket."""
        self._transport, _ = await self._loop.create_datagram_endpoint(
            asyncio.DatagramProtocol,
            family=socket.AF_INET,
            remote_addr=self._addr
        )

    def _send(self, data: str) -> None:
        """Start an async task to send data to statsd."""
        scheduling.create_task(self._async_send(data), event_loop=self._loop)

    async def _async_send(self, data: str) -> None:
        """Send data to the statsd server using the async transport."""
        self._transport.sendto(data.encode("ascii"), self._addr)


__all__ = ["AsyncStatsClient"]



import asyncio
import socket
import types
import warnings
from contextlib import suppress

import aiohttp
import discord
from discord import app_commands
from discord.ext import commands

from pydis_core.async_stats import AsyncStatsClient
from pydis_core.site_api import APIClient
from pydis_core.utils import scheduling
from pydis_core.utils._extensions import walk_extensions
from pydis_core.utils.error_handling.commands import CommandErrorManager
from pydis_core.utils.logging import get_logger

try:
    from async_rediscache import RedisSession
    REDIS_AVAILABLE = True
except ImportError:
    RedisSession = object
    REDIS_AVAILABLE = False

log = get_logger()


class StartupError(Exception):
    """Exception class for startup errors."""

    def __init__(self, base: Exception):
        super().__init__()
        self.exception = base


class CommandTreeBase(app_commands.CommandTree):
    """A sub-class of the Command tree that implements common features that Python Discord bots use."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Instance is None since discordpy only passes an instance of the client to the command tree in its constructor.
        self.command_error_manager: CommandErrorManager | None = None

    async def on_error(self, interaction: discord.Interaction, error: app_commands.AppCommandError) -> None:
        """A callback that is called when any command raises an :exc:`AppCommandError`."""
        if not self.command_error_manager:
            log.warning("Command error manager hasn't been loaded in the command tree.")
            await super().on_error(interaction, error)
            return
        await self.command_error_manager.handle_error(error, interaction)


class BotBase(commands.Bot):
    """
    A sub-class that implements many common features that Python Discord bots use.

    Attributes:
        guild_id (int): ID of the guild that the bot belongs to.
        http_session (aiohttp.ClientSession): The http session used for sending out HTTP requests.
        api_client (pydis_core.site_api.APIClient): The API client used for communications with the site service.
        statsd_url (str): The url that statsd sends metrics to.
        redis_session (async_rediscache.RedisSession): The redis session used to communicate with the Redis instance.
        stats (pydis_core.async_stats.AsyncStatsClient): The statsd client that sends metrics.
        all_extensions (frozenset[str]): All extensions that were found within the ``module`` passed to
            ``self.load_extensions``. Use ``self.extensions`` to get the loaded extensions.

    """

    def __init__(
        self,
        *args,
        guild_id: int,
        allowed_roles: list,
        http_session: aiohttp.ClientSession,
        redis_session: RedisSession | None = None,
        api_client: APIClient | None = None,
        statsd_url: str | None = None,
        **kwargs,
    ):
        """
        Initialise the base bot instance.

        Args:
            guild_id: The ID of the guild used for :func:`wait_until_guild_available`.
            allowed_roles: A list of role IDs that the bot is allowed to mention.
            http_session (aiohttp.ClientSession): The session to use for the bot.
            redis_session: The `async_rediscache.RedisSession`_ to use for the bot.
            api_client: The :obj:`pydis_core.site_api.APIClient` instance to use for the bot.
            statsd_url: The URL of the statsd server to use for the bot. If not given,
                a dummy statsd client will be created.

        .. _async_rediscache.RedisSession: https://github.com/SebastiaanZ/async-rediscache#creating-a-redissession
        """
        super().__init__(
            *args,
            allowed_roles=allowed_roles,
            tree_cls=CommandTreeBase,
            **kwargs,
        )

        self.command_error_manager: CommandErrorManager | None = None
        self.guild_id = guild_id
        self.http_session = http_session
        self.api_client = api_client
        self.statsd_url = statsd_url

        if redis_session and not REDIS_AVAILABLE:
            warnings.warn("redis_session kwarg passed, but async-rediscache not installed!", stacklevel=2)
        elif redis_session:
            self.redis_session = redis_session

        self._resolver: aiohttp.AsyncResolver | None = None
        self._connector: aiohttp.TCPConnector | None = None

        self._statsd_timerhandle: asyncio.TimerHandle | None = None
        self._guild_available: asyncio.Event | None = None
        self._extension_loading_task: asyncio.Task | None = None

        self.stats: AsyncStatsClient | None = None

        self.all_extensions: frozenset[str] | None = None

    def register_command_error_manager(self, manager: CommandErrorManager) -> None:
        """
        Bind an instance of the command error manager to both the bot and the command tree.

        The reason this doesn't happen in the constructor is because error handlers might need an instance of the bot.
        So registration needs to happen once the bot instance has been created.
        """
        self.command_error_manager = manager
        self.tree.command_error_manager = manager

    def _connect_statsd(
        self,
        statsd_url: str,
        loop: asyncio.AbstractEventLoop,
        retry_after: int = 2,
        attempt: int = 1
    ) -> None:
        """Callback used to retry a connection to statsd if it should fail."""
        if attempt >= 8:
            log.error(
                "Reached 8 attempts trying to reconnect AsyncStatsClient to %s. "
                "Aborting and leaving the dummy statsd client in place.",
                statsd_url,
            )
            return

        try:
            self.stats = AsyncStatsClient(loop, statsd_url, 8125, prefix="bot")
        except socket.gaierror:
            log.warning(f"Statsd client failed to connect (Attempt(s): {attempt})")
            # Use a fallback strategy for retrying, up to 8 times.
            self._statsd_timerhandle = loop.call_later(
                retry_after,
                self._connect_statsd,
                statsd_url,
                retry_after * 2,
                attempt + 1
            )

    async def _load_extensions(self, module: types.ModuleType) -> None:
        """Load all the extensions within the given module and save them to ``self.all_extensions``."""
        log.info("Waiting for guild %d to be available before loading extensions.", self.guild_id)

        await self.wait_until_guild_available()
        log.info("Loading extensions...")
        self.all_extensions = walk_extensions(module)

        for extension in self.all_extensions:
            scheduling.create_task(self.load_extension(extension))

    async def _sync_app_commands(self) -> None:
        """Sync global & guild specific application commands after extensions are loaded."""
        await self._extension_loading_task
        await self.tree.sync()
        await self.tree.sync(guild=discord.Object(self.guild_id))

    async def load_extensions(self, module: types.ModuleType, *, sync_app_commands: bool = True) -> None:
        """
        Load all the extensions within the given ``module`` and save them to ``self.all_extensions``.

        Args:
            sync_app_commands: Whether to sync app commands after all extensions are loaded.
        """
        self._extension_loading_task = scheduling.create_task(self._load_extensions(module))
        if sync_app_commands:
            scheduling.create_task(self._sync_app_commands())

    def _add_root_aliases(self, command: commands.Command) -> None:
        """Recursively add root aliases for ``command`` and any of its subcommands."""
        if isinstance(command, commands.Group):
            for subcommand in command.commands:
                self._add_root_aliases(subcommand)

        for alias in getattr(command, "root_aliases", ()):
            if alias in self.all_commands:
                raise commands.CommandRegistrationError(alias, alias_conflict=True)

            self.all_commands[alias] = command

    def _remove_root_aliases(self, command: commands.Command) -> None:
        """Recursively remove root aliases for ``command`` and any of its subcommands."""
        if isinstance(command, commands.Group):
            for subcommand in command.commands:
                self._remove_root_aliases(subcommand)

        for alias in getattr(command, "root_aliases", ()):
            self.all_commands.pop(alias, None)

    async def add_cog(self, cog: commands.Cog) -> None:
        """Add the given ``cog`` to the bot and log the operation."""
        await super().add_cog(cog)
        log.info(f"Cog loaded: {cog.qualified_name}")

    def add_command(self, command: commands.Command) -> None:
        """Add ``command`` as normal and then add its root aliases to the bot."""
        super().add_command(command)
        self._add_root_aliases(command)

    def remove_command(self, name: str) -> commands.Command | None:
        """
        Remove a command/alias as normal and then remove its root aliases from the bot.

        Individual root aliases cannot be removed by this function.
        To remove them, either remove the entire command or manually edit `bot.all_commands`.
        """
        command = super().remove_command(name)
        if command is None:
            # Even if it's a root alias, there's no way to get the Bot instance to remove the alias.
            return None

        self._remove_root_aliases(command)
        return command

    def clear(self) -> None:
        """Not implemented! Re-instantiate the bot instead of attempting to re-use a closed one."""
        raise NotImplementedError("Re-using a Bot object after closing it is not supported.")

    async def on_guild_unavailable(self, guild: discord.Guild) -> None:
        """Clear the internal guild available event when self.guild_id becomes unavailable."""
        if guild.id != self.guild_id:
            return

        self._guild_available.clear()

    async def on_guild_available(self, guild: discord.Guild) -> None:
        """
        Set the internal guild available event when self.guild_id becomes available.

        If the cache appears to still be empty (no members, no channels, or no roles), the event
        will not be set and `guild_available_but_cache_empty` event will be emitted.
        """
        if guild.id != self.guild_id:
            return

        if not guild.roles or not guild.members or not guild.channels:
            msg = "Guild available event was dispatched but the cache appears to still be empty!"
            await self.log_to_dev_log(msg)
            return

        self._guild_available.set()

    async def log_to_dev_log(self, message: str) -> None:
        """Log the given message to #dev-log."""

    async def wait_until_guild_available(self) -> None:
        """
        Wait until the guild that matches the ``guild_id`` given at init is available (and the cache is ready).

        The on_ready event is inadequate because it only waits 2 seconds for a GUILD_CREATE
        gateway event before giving up and thus not populating the cache for unavailable guilds.
        """
        await self._guild_available.wait()

    async def process_commands(self, message: discord.Message) -> None:
        """
        Overwrite default Discord.py behaviour to process commands only after ensuring extensions are loaded.

        This extension check is only relevant for clients that make use of :obj:`pydis_core.BotBase.load_extensions`.
        """
        if self._extension_loading_task:
            await self._extension_loading_task
        await super().process_commands(message)

    async def setup_hook(self) -> None:
        """
        An async init to startup generic services.

        Connects to statsd, and calls
        :func:`AsyncStatsClient.create_socket <pydis_core.async_stats.AsyncStatsClient.create_socket>`
        and :func:`ping_services`.
        """
        loop = asyncio.get_running_loop()

        self._guild_available = asyncio.Event()

        self._resolver = aiohttp.AsyncResolver()
        self._connector = aiohttp.TCPConnector(
            resolver=self._resolver,
            family=socket.AF_INET,
        )
        self.http.connector = self._connector

        if getattr(self, "redis_session", False) and not self.redis_session.valid:
            # If the RedisSession was somehow closed, we try to reconnect it
            # here. Normally, this shouldn't happen.
            await self.redis_session.connect(ping=True)

        # Create dummy stats client first, in case `statsd_url` is unreachable or None
        self.stats = AsyncStatsClient(loop, "127.0.0.1")
        if self.statsd_url:
            self._connect_statsd(self.statsd_url, loop)

        await self.stats.create_socket()

        try:
            await self.ping_services()
        except Exception as e:  # noqa: BLE001
            raise StartupError(e)

    async def ping_services(self) -> None:
        """Ping all required services on setup to ensure they are up before starting."""

    async def close(self) -> None:
        """Close the Discord connection, and the aiohttp session, connector, statsd client, and resolver."""
        # Done before super().close() to allow tasks finish before the HTTP session closes.
        for ext in list(self.extensions):
            with suppress(Exception):
                await self.unload_extension(ext)

        for cog in list(self.cogs):
            with suppress(Exception):
                await self.remove_cog(cog)

        # Now actually do full close of bot
        await super().close()

        if self.api_client:
            await self.api_client.close()

        if self.http_session:
            await self.http_session.close()

        if self._connector:
            await self._connector.close()

        if self._resolver:
            await self._resolver.close()

        if getattr(self.stats, "_transport", False):
            self.stats._transport.close()  # noqa: SLF001

        if self._statsd_timerhandle:
            self._statsd_timerhandle.cancel()



import asyncio
import inspect
import types
from collections import defaultdict
from collections.abc import Awaitable, Callable, Hashable
from functools import partial
from typing import Any
from weakref import WeakValueDictionary

from pydis_core.utils import function
from pydis_core.utils.function import command_wraps
from pydis_core.utils.logging import get_logger

log = get_logger(__name__)
__lock_dicts = defaultdict(WeakValueDictionary)

_IdCallableReturn = Hashable | Awaitable[Hashable]
_IdCallable = Callable[[function.BoundArgs], _IdCallableReturn]
ResourceId = Hashable | _IdCallable


class LockedResourceError(RuntimeError):
    """
    Exception raised when an operation is attempted on a locked resource.

    Attributes:
        type (str): Name of the locked resource's type
        id (typing.Hashable): ID of the locked resource
    """

    def __init__(self, resource_type: str, resource_id: Hashable):
        self.type = resource_type
        self.id = resource_id

        super().__init__(
            f"Cannot operate on {self.type.lower()} `{self.id}`; "
            "it is currently locked and in use by another operation."
        )


class SharedEvent:
    """
    Context manager managing an internal event exposed through the wait coro.

    While any code is executing in this context manager, the underlying event will not be set;
    when all of the holders finish the event will be set.
    """

    def __init__(self):
        self._active_count = 0
        self._event = asyncio.Event()
        self._event.set()

    def __enter__(self):
        """Increment the count of the active holders and clear the internal event."""
        self._active_count += 1
        self._event.clear()

    def __exit__(self, _exc_type, _exc_val, _exc_tb):  # noqa: ANN001
        """Decrement the count of the active holders; if 0 is reached set the internal event."""
        self._active_count -= 1
        if not self._active_count:
            self._event.set()

    async def wait(self) -> None:
        """Wait for all active holders to exit."""
        await self._event.wait()


def lock(
    namespace: Hashable,
    resource_id: ResourceId,
    *,
    raise_error: bool = False,
    wait: bool = False,
) -> Callable:
    """
    Turn the decorated coroutine function into a mutually exclusive operation on a `resource_id`.

    If decorating a command, this decorator must go before (below) the `command` decorator.

    Arguments:
        namespace (typing.Hashable): An identifier used to prevent collisions among resource IDs.
        resource_id: identifies a resource on which to perform a mutually exclusive operation.
            It may also be a callable or awaitable which will return the resource ID given an ordered
            mapping of the parameters' names to arguments' values.
        raise_error (bool): If True, raise `LockedResourceError` if the lock cannot be acquired.
        wait (bool): If True, wait until the lock becomes available. Otherwise, if any other mutually
            exclusive function currently holds the lock for a resource, do not run the decorated function
            and return None.

    Raises:
        :exc:`LockedResourceError`: If the lock can't be acquired and `raise_error` is set to True.
    """
    def decorator(func: types.FunctionType) -> types.FunctionType:
        name = func.__name__

        @command_wraps(func)
        async def wrapper(*args, **kwargs) -> Any:
            log.trace(f"{name}: mutually exclusive decorator called")

            if callable(resource_id):
                log.trace(f"{name}: binding args to signature")
                bound_args = function.get_bound_args(func, args, kwargs)

                log.trace(f"{name}: calling the given callable to get the resource ID")
                id_ = resource_id(bound_args)

                if inspect.isawaitable(id_):
                    log.trace(f"{name}: awaiting to get resource ID")
                    id_ = await id_
            else:
                id_ = resource_id

            log.trace(f"{name}: getting the lock object for resource {namespace!r}:{id_!r}")

            # Get the lock for the ID. Create a lock if one doesn't exist yet.
            locks = __lock_dicts[namespace]
            lock_ = locks.setdefault(id_, asyncio.Lock())

            # It's safe to check an asyncio.Lock is free before acquiring it because:
            #   1. Synchronous code like `if not lock_.locked()` does not yield execution
            #   2. `asyncio.Lock.acquire()` does not internally await anything if the lock is free
            #   3. awaits only yield execution to the event loop at actual I/O boundaries
            if wait or not lock_.locked():
                log.debug(f"{name}: acquiring lock for resource {namespace!r}:{id_!r}...")
                async with lock_:
                    return await func(*args, **kwargs)
            else:
                log.info(f"{name}: aborted because resource {namespace!r}:{id_!r} is locked")
                if raise_error:
                    raise LockedResourceError(str(namespace), id_)
                return None

        return wrapper
    return decorator


def lock_arg(
    namespace: Hashable,
    name_or_pos: function.Argument,
    func: Callable[[Any], _IdCallableReturn] | None = None,
    *,
    raise_error: bool = False,
    wait: bool = False,
) -> Callable:
    """
    Apply the `lock` decorator using the value of the arg at the given name/position as the ID.

    See `lock` docs for more information.

    Arguments:
        func: An optional callable or awaitable which will return the ID given the argument value.
    """
    decorator_func = partial(lock, namespace, raise_error=raise_error, wait=wait)
    return function.get_arg_value_wrapper(decorator_func, name_or_pos, func)



from collections.abc import Sequence
from contextlib import suppress
from functools import partial

import discord
from discord.abc import User
from discord.ext.commands import Context, Paginator
from pydantic import BaseModel

from pydis_core.utils.logging import get_logger
from pydis_core.utils.messages import reaction_check

log = get_logger(__name__)


class PaginationEmojis(BaseModel):
    """The emojis that will be used for pagination."""

    first: str = "\u23EE"
    left: str = "\u2B05"
    right: str = "\u27A1"
    last: str = "\u23ED"
    delete: str = "<:trashcan:637136429717389331>"


class EmptyPaginatorEmbedError(Exception):
    """Raised when attempting to paginate with empty contents."""


class LinePaginator(Paginator):
    """
    A class that aids in paginating code blocks for Discord messages.

    Args:
        pagination_emojis (PaginationEmojis): The emojis used to navigate pages.
        prefix (str): The prefix inserted to every page. e.g. three backticks.
        suffix (str): The suffix appended at the end of every page. e.g. three backticks.
        max_size (int): The maximum amount of codepoints allowed in a page.
        scale_to_size (int): The maximum amount of characters a single line can scale up to.
        max_lines (int): The maximum amount of lines allowed in a page.
    """

    def __init__(
        self,
        prefix: str = "```",
        suffix: str = "```",
        max_size: int = 4000,
        scale_to_size: int = 4000,
        max_lines: int | None = None,
        linesep: str = "\n"
    ) -> None:
        """
        This function overrides the Paginator.__init__ from inside discord.ext.commands.

        It overrides in order to allow us to configure the maximum number of lines per page.
        """
        # Embeds that exceed 4096 characters will result in an HTTPException
        # (Discord API limit), so we've set a limit of 4000
        if max_size > 4000:
            raise ValueError(f"max_size must be <= 4,000 characters. ({max_size} > 4000)")

        super().__init__(
            prefix,
            suffix,
            max_size - len(suffix),
            linesep
        )

        if scale_to_size < max_size:
            raise ValueError(f"scale_to_size must be >= max_size. ({scale_to_size} < {max_size})")

        if scale_to_size > 4000:
            raise ValueError(f"scale_to_size must be <= 4,000 characters. ({scale_to_size} > 4000)")

        self.scale_to_size = scale_to_size - len(suffix)
        self.max_lines = max_lines
        self._current_page = [prefix]
        self._linecount = 0
        self._count = len(prefix) + 1  # prefix + newline
        self._pages = []

    def add_line(self, line: str = "", *, empty: bool = False) -> None:
        """
        Adds a line to the current page.

        If a line on a page exceeds `max_size` characters, then `max_size` will go up to
        `scale_to_size` for a single line before creating a new page for the overflow words. If it
        is still exceeded, the excess characters are stored and placed on the next pages unti
        there are none remaining (by word boundary). The line is truncated if `scale_to_size` is
        still exceeded after attempting to continue onto the next page.

        In the case that the page already contains one or more lines and the new lines would cause
        `max_size` to be exceeded, a new page is created. This is done in order to make a best
        effort to avoid breaking up single lines across pages, while keeping the total length of the
        page at a reasonable size.

        This function overrides the `Paginator.add_line` from inside `discord.ext.commands`.

        It overrides in order to allow us to configure the maximum number of lines per page.

        Args:
            line (str): The line to add to the paginated content.
            empty (bool): Indicates whether an empty line should be added at the end.
        """
        remaining_words = None
        if len(line) > (max_chars := self.max_size - len(self.prefix) - 2):
            if len(line) > self.scale_to_size:
                line, remaining_words = self._split_remaining_words(line, max_chars)
                if len(line) > self.scale_to_size:
                    log.debug("Could not continue to next page, truncating line.")
                    line = line[:self.scale_to_size]

        # Check if we should start a new page or continue the line on the current one
        if self.max_lines is not None and self._linecount >= self.max_lines:
            log.debug("max_lines exceeded, creating new page.")
            self._new_page()
        elif self._count + len(line) + 1 > self.max_size and self._linecount > 0:
            log.debug("max_size exceeded on page with lines, creating new page.")
            self._new_page()

        self._linecount += 1

        self._count += len(line) + 1
        self._current_page.append(line)

        if empty:
            self._current_page.append("")
            self._count += 1

        # Start a new page if there were any overflow words
        if remaining_words:
            self._new_page()
            self.add_line(remaining_words)

    def _new_page(self) -> None:
        """
        Internal: start a new page for the paginator.

        This closes the current page and resets the counters for the new page's line count and
        character count.
        """
        self._linecount = 0
        self._count = len(self.prefix) + 1
        self.close_page()

    def _split_remaining_words(self, line: str, max_chars: int) -> tuple[str, str | None]:
        """
        Internal: split a line into two strings -- reduced_words and remaining_words.

        reduced_words: the remaining words in `line`, after attempting to remove all words that
            exceed `max_chars` (rounding down to the nearest word boundary).

        remaining_words: the words in `line` which exceed `max_chars`. This value is None if
            no words could be split from `line`.

        If there are any remaining_words, an ellipses is appended to reduced_words and a
        continuation header is inserted before remaining_words to visually communicate the line
        continuation.

        Return a tuple in the format (reduced_words, remaining_words).
        """
        reduced_words = []
        remaining_words = []

        # "(Continued)" is used on a line by itself to indicate the continuation of last page
        continuation_header = "(Continued)\n-----------\n"
        reduced_char_count = 0
        is_full = False

        for word in line.split(" "):
            if not is_full:
                if len(word) + reduced_char_count <= max_chars:
                    reduced_words.append(word)
                    reduced_char_count += len(word) + 1
                else:
                    # If reduced_words is empty, we were unable to split the words across pages
                    if not reduced_words:
                        return line, None
                    is_full = True
                    remaining_words.append(word)
            else:
                remaining_words.append(word)

        return (
            " ".join(reduced_words) + "..." if remaining_words else "",
            continuation_header + " ".join(remaining_words) if remaining_words else None
        )

    @classmethod
    async def paginate(
        cls,
        pagination_emojis: PaginationEmojis,
        lines: list[str],
        ctx: Context | discord.Interaction,
        embed: discord.Embed,
        *,
        prefix: str = "",
        suffix: str = "",
        max_lines: int | None = None,
        max_size: int = 500,
        scale_to_size: int = 4000,
        empty: bool = True,
        restrict_to_user: User | None = None,
        timeout: int = 300,  # noqa: ASYNC109
        footer_text: str | None = None,
        url: str | None = None,
        exception_on_empty_embed: bool = False,
        reply: bool = False,
        allowed_roles: Sequence[int] | None = None,
    ) -> discord.Message | None:
        """
        Use a paginator and set of reactions to provide pagination over a set of lines.

        The reactions are used to switch page, or to finish with pagination.

        When used, this will send a message using `ctx.send()` and apply a set of reactions to it. These reactions may
        be used to change page, or to remove pagination from the message.

        Pagination will also be removed automatically if no reaction is added for five minutes (300 seconds).

        The interaction will be limited to `restrict_to_user` (ctx.author by default) or
        to any user with a moderation role.

        Args:
            pagination_emojis (PaginationEmojis): The emojis used to navigate pages.
            lines (list[str]): A list of lines to be added to the paginated content.
            ctx (:obj:`discord.ext.commands.Context`): The context in which the pagination is needed.
            embed (:obj:`discord.Embed`): The embed that holds the content, it serves as the page.
            prefix (str): The prefix inserted to every page. e.g. three backticks.
            suffix (str): The suffix appended at the end of every page. e.g. three backticks.
            max_lines (int): The maximum amount of lines allowed in a page.
            max_size (int): The maximum amount of codepoints allowed in a page.
            scale_to_size (int): The maximum amount of characters a single line can scale up to.
            empty (bool): Indicates whether an empty line should be added to each provided line.
            restrict_to_user (:obj:`discord.User`): The user to whom interaction with the pages should be restricted.
            timeout (int): The timeout after which users cannot change pages anymore.
            footer_text (str): Text to be added as a footer for each page.
            url (str): The url to be set for the pagination embed.
            exception_on_empty_embed (bool): Indicates whether to raise an exception when no lines are provided.
            reply (bool): Indicates whether to send the page as a reply to the context's message.
            allowed_roles (Sequence[int]): A list of role ids that are allowed to change pages.

        Example:
        >>> embed = discord.Embed()
        >>> embed.set_author(name="Some Operation", url=url, icon_url=icon)
        >>> await LinePaginator.paginate(pagination_emojis, [line for line in lines], ctx, embed)
        """
        paginator = cls(prefix=prefix, suffix=suffix, max_size=max_size,
                        max_lines=max_lines, scale_to_size=scale_to_size)
        current_page = 0

        if not restrict_to_user:
            if isinstance(ctx, discord.Interaction):
                restrict_to_user = ctx.user
            else:
                restrict_to_user = ctx.author

        if not lines:
            if exception_on_empty_embed:
                log.exception("Pagination asked for empty lines iterable")
                raise EmptyPaginatorEmbedError("No lines to paginate")

            log.debug("No lines to add to paginator, adding '(nothing to display)' message")
            lines.append("*(nothing to display)*")

        for line in lines:
            try:
                paginator.add_line(line, empty=empty)
            except Exception:
                log.exception(f"Failed to add line to paginator: '{line}'")
                raise  # Should propagate
            else:
                log.trace(f"Added line to paginator: '{line}'")

        log.debug(f"Paginator created with {len(paginator.pages)} pages")

        embed.description = paginator.pages[current_page]

        reference = ctx.message if reply else None

        if len(paginator.pages) <= 1:
            if footer_text:
                embed.set_footer(text=footer_text)
                log.trace(f"Setting embed footer to '{footer_text}'")

            if url:
                embed.url = url
                log.trace(f"Setting embed url to '{url}'")

            log.debug("There's less than two pages, so we won't paginate - sending single page on its own")

            if isinstance(ctx, discord.Interaction):
                return await ctx.response.send_message(embed=embed)
            return await ctx.send(embed=embed, reference=reference)

        if footer_text:
            embed.set_footer(text=f"{footer_text} (Page {current_page + 1}/{len(paginator.pages)})")
        else:
            embed.set_footer(text=f"Page {current_page + 1}/{len(paginator.pages)}")
        log.trace(f"Setting embed footer to '{embed.footer.text}'")

        if url:
            embed.url = url
            log.trace(f"Setting embed url to '{url}'")

        log.debug("Sending first page to channel...")

        if isinstance(ctx, discord.Interaction):
            await ctx.response.send_message(embed=embed)
            message = await ctx.original_response()
        else:
            message = await ctx.send(embed=embed, reference=reference)

        log.debug("Adding emoji reactions to message...")

        pagination_emoji = list(pagination_emojis.model_dump().values())

        for emoji in pagination_emoji:
            # Add all the applicable emoji to the message
            log.trace(f"Adding reaction: {emoji!r}")
            await message.add_reaction(emoji)

        check = partial(
            reaction_check,
            message_id=message.id,
            allowed_emoji=pagination_emoji,
            allowed_users=(restrict_to_user.id,),
            allowed_roles=allowed_roles,
        )

        while True:
            try:
                if isinstance(ctx, discord.Interaction):
                    reaction, user = await ctx.client.wait_for("reaction_add", timeout=timeout, check=check)
                else:
                    reaction, user = await ctx.bot.wait_for("reaction_add", timeout=timeout, check=check)
                log.trace(f"Got reaction: {reaction}")
            except TimeoutError:
                log.debug("Timed out waiting for a reaction")
                break  # We're done, no reactions for the last 5 minutes

            if str(reaction.emoji) == pagination_emojis.delete:
                log.debug("Got delete reaction")
                return await message.delete()
            if reaction.emoji in pagination_emoji:
                total_pages = len(paginator.pages)
                try:
                    await message.remove_reaction(reaction.emoji, user)
                except discord.HTTPException as e:
                    # Suppress if trying to act on an archived thread.
                    if e.code != 50083:
                        raise

                if reaction.emoji == pagination_emojis.first:
                    current_page = 0
                    log.debug(f"Got first page reaction - changing to page 1/{total_pages}")
                elif reaction.emoji == pagination_emojis.last:
                    current_page = len(paginator.pages) - 1
                    log.debug(f"Got last page reaction - changing to page {current_page + 1}/{total_pages}")
                elif reaction.emoji == pagination_emojis.left:
                    if current_page <= 0:
                        log.debug("Got previous page reaction, but we're on the first page - ignoring")
                        continue

                    current_page -= 1
                    log.debug(f"Got previous page reaction - changing to page {current_page + 1}/{total_pages}")
                elif reaction.emoji == pagination_emojis.right:
                    if current_page >= len(paginator.pages) - 1:
                        log.debug("Got next page reaction, but we're on the last page - ignoring")
                        continue

                    current_page += 1
                    log.debug(f"Got next page reaction - changing to page {current_page + 1}/{total_pages}")

                embed.description = paginator.pages[current_page]

                if footer_text:
                    embed.set_footer(text=f"{footer_text} (Page {current_page + 1}/{len(paginator.pages)})")
                else:
                    embed.set_footer(text=f"Page {current_page + 1}/{len(paginator.pages)}")

                try:
                    await message.edit(embed=embed)
                except discord.HTTPException as e:
                    if e.code == 50083:
                        # Trying to act on an archived thread, just ignore and abort
                        break
                    raise

        log.debug("Ending pagination and clearing reactions.")
        with suppress(discord.NotFound):
            try:
                await message.clear_reactions()
            except discord.HTTPException as e:
                # Suppress if trying to act on an archived thread.
                if e.code != 50083:
                    raise



"""Common logging related functions."""

import logging
import typing

if typing.TYPE_CHECKING:
    LoggerClass = logging.Logger
else:
    LoggerClass = logging.getLoggerClass()

TRACE_LEVEL = 5
log_format = logging.Formatter("%(asctime)s | %(levelname)s | %(name)s | %(message)s")
"""
A :obj:`logging.Formatter` that can be used to standardise logging across all projects using pydis_core.

:meta hide-value:
"""


class CustomLogger(LoggerClass):
    """Custom implementation of the :obj:`logging.Logger` class with an added :obj:`trace` method."""

    def trace(self, msg: str, *args, **kwargs) -> None:
        """
        Log the given message with the severity ``"TRACE"``.

        To pass exception information, use the keyword argument exc_info with a true value:

        .. code-block:: py

            logger.trace("Houston, we have an %s", "interesting problem", exc_info=1)

        Args:
            msg: The message to be logged.
            args, kwargs: Passed to the base log function as is.
        """
        if self.isEnabledFor(TRACE_LEVEL):
            self.log(TRACE_LEVEL, msg, *args, **kwargs)


def get_logger(name: str | None = None) -> CustomLogger:
    """
    Utility to make mypy recognise that logger is of type :obj:`CustomLogger`.

    Args:
        name: The name given to the logger.

    Returns:
        An instance of the :obj:`CustomLogger` class.
    """
    return typing.cast(CustomLogger, logging.getLogger(name))


# Setup trace level logging so that we can use it within pydis_core.
logging.TRACE = TRACE_LEVEL
logging.setLoggerClass(CustomLogger)
logging.addLevelName(TRACE_LEVEL, "TRACE")

get_logger("discord").setLevel(logging.WARNING)
get_logger("websockets").setLevel(logging.WARNING)
get_logger("chardet").setLevel(logging.WARNING)
get_logger("async_rediscache").setLevel(logging.WARNING)
# Set back to the default of INFO even if asyncio's debug mode is enabled.
get_logger("asyncio").setLevel(logging.INFO)



"""Common regular expressions."""

import re

DISCORD_INVITE = re.compile(
    r"(https?:\/\/)?(discord:\/*)?"              # Optional protocols
    r"(www\.)?"                                  # Optional www
    r"[@#]*"                                     # Optional @ or # symbols
    r"(\B|discord(app)?)"                        # Optional discord(app)
    r"([.,]|dot)"                                # Various characters to cover dots
    r"("
        r"(gg|me)"                               # TLDs that embed within discord
        r"|com(\/|slash|\\)invite"               # Only match com/invite
    r")"
    r"(/|slash|\\+)"                             # / or 'slash' or 1+ of \
    r"(?P<invite>\S+)",                          # the invite code itself
    flags=re.IGNORECASE
)
"""
Regex for Discord server invites.

.. warning::
    This regex pattern will capture until a whitespace, if you are to use the 'invite' capture group in
    any HTTP requests or similar. Please ensure you sanitise the output using something
    such as :func:`urllib.parse.quote`.

:meta hide-value:
"""

FORMATTED_CODE_REGEX = re.compile(
    r"(?P<delim>(?P<block>```)|``?)"        # code delimiter: 1-3 backticks; (?P=block) only matches if it's a block
    r"(?(block)(?:(?P<lang>[a-z]+)\n)?)"    # if we're in a block, match optional language (only letters plus newline)
    r"(?:[ \t]*\n)*"                        # any blank (empty or tabs/spaces only) lines before the code
    r"(?P<code>.*?)"                        # extract all code inside the markup
    r"\s*"                                  # any more whitespace before the end of the code markup
    r"(?P=delim)",                          # match the exact same delimiter from the start again
    flags=re.DOTALL | re.IGNORECASE         # "." also matches newlines, case insensitive
)
"""
Regex for formatted code, using Discord's code blocks.

:meta hide-value:
"""

RAW_CODE_REGEX = re.compile(
    r"^(?:[ \t]*\n)*"                       # any blank (empty or tabs/spaces only) lines before the code
    r"(?P<code>.*?)"                        # extract all the rest as code
    r"\s*$",                                # any trailing whitespace until the end of the string
    flags=re.DOTALL                         # "." also matches newlines
)
"""
Regex for raw code, *not* using Discord's code blocks.

:meta hide-value:
"""



"""Generic python scheduler."""

import asyncio
import contextlib
import inspect
import typing
from collections import abc
from datetime import UTC, datetime
from functools import partial

from discord.errors import Forbidden

from pydis_core.utils import logging
from pydis_core.utils.error_handling import handle_forbidden_from_block

_background_tasks: set[asyncio.Task] = set()


class Scheduler:
    """
    Schedule the execution of coroutines and keep track of them.

    When instantiating a :obj:`Scheduler`, a name must be provided. This name is used to distinguish the
    instance's log messages from other instances. Using the name of the class or module containing
    the instance is suggested.

    Coroutines can be scheduled immediately with :obj:`schedule` or in the future with :obj:`schedule_at`
    or :obj:`schedule_later`. A unique ID is required to be given in order to keep track of the
    resulting Tasks. Any scheduled task can be cancelled prematurely using :obj:`cancel` by providing
    the same ID used to schedule it.

    The ``in`` operator is supported for checking if a task with a given ID is currently scheduled.

    Any exception raised in a scheduled task is logged when the task is done.
    """

    def __init__(self, name: str):
        """
        Initialize a new :obj:`Scheduler` instance.

        Args:
            name: The name of the :obj:`Scheduler`. Used in logging, and namespacing.
        """
        self.name = name

        self._log = logging.get_logger(f"{__name__}.{name}")
        self._scheduled_tasks: dict[abc.Hashable, asyncio.Task] = {}

    def __contains__(self, task_id: abc.Hashable) -> bool:
        """
        Return :obj:`True` if a task with the given ``task_id`` is currently scheduled.

        Args:
            task_id: The task to look for.

        Returns:
            :obj:`True` if the task was found.
        """
        return task_id in self._scheduled_tasks

    def schedule(self, task_id: abc.Hashable, coroutine: abc.Coroutine) -> None:
        """
        Schedule the execution of a ``coroutine``.

        If a task with ``task_id`` already exists, close ``coroutine`` instead of scheduling it. This
        prevents unawaited coroutine warnings. Don't pass a coroutine that'll be re-used elsewhere.

        Args:
            task_id: A unique ID to create the task with.
            coroutine: The function to be called.
        """
        self._log.trace(f"Scheduling task #{task_id}...")

        msg = f"Cannot schedule an already started coroutine for #{task_id}"
        if inspect.getcoroutinestate(coroutine) != "CORO_CREATED":
            raise ValueError(msg)

        if task_id in self._scheduled_tasks:
            self._log.debug(f"Did not schedule task #{task_id}; task was already scheduled.")
            coroutine.close()
            return

        task = asyncio.create_task(_coro_wrapper(coroutine), name=f"{self.name}_{task_id}")
        task.add_done_callback(partial(self._task_done_callback, task_id))

        self._scheduled_tasks[task_id] = task
        self._log.debug(f"Scheduled task #{task_id} {id(task)}.")

    def schedule_at(self, time: datetime, task_id: abc.Hashable, coroutine: abc.Coroutine) -> None:
        """
        Schedule ``coroutine`` to be executed at the given ``time``.

        If ``time`` is timezone aware, then use that timezone to calculate now() when subtracting.
        If ``time`` is naïve, then use UTC.

        If ``time`` is in the past, schedule ``coroutine`` immediately.

        If a task with ``task_id`` already exists, close ``coroutine`` instead of scheduling it. This
        prevents unawaited coroutine warnings. Don't pass a coroutine that'll be re-used elsewhere.

        Args:
            time: The time to start the task.
            task_id: A unique ID to create the task with.
            coroutine: The function to be called.
        """
        now_datetime = datetime.now(time.tzinfo) if time.tzinfo else datetime.now(tz=UTC)
        delay = (time - now_datetime).total_seconds()
        if delay > 0:
            coroutine = self._await_later(delay, task_id, coroutine)

        self.schedule(task_id, coroutine)

    def schedule_later(
        self,
        delay: float,
        task_id: abc.Hashable,
        coroutine: abc.Coroutine
    ) -> None:
        """
        Schedule ``coroutine`` to be executed after ``delay`` seconds.

        If a task with ``task_id`` already exists, close ``coroutine`` instead of scheduling it. This
        prevents unawaited coroutine warnings. Don't pass a coroutine that'll be re-used elsewhere.

        Args:
            delay: How long to wait before starting the task.
            task_id: A unique ID to create the task with.
            coroutine: The function to be called.
        """
        self.schedule(task_id, self._await_later(delay, task_id, coroutine))

    def cancel(self, task_id: abc.Hashable) -> None:
        """
        Unschedule the task identified by ``task_id``. Log a warning if the task doesn't exist.

        Args:
            task_id: The task's unique ID.
        """
        self._log.trace(f"Cancelling task #{task_id}...")

        try:
            task = self._scheduled_tasks.pop(task_id)
        except KeyError:
            self._log.warning(f"Failed to unschedule {task_id} (no task found).")
        else:
            task.cancel()

            self._log.debug(f"Unscheduled task #{task_id} {id(task)}.")

    def cancel_all(self) -> None:
        """Unschedule all known tasks."""
        self._log.debug("Unscheduling all tasks")

        for task_id in self._scheduled_tasks.copy():
            self.cancel(task_id)

    async def _await_later(
        self,
        delay: float,
        task_id: abc.Hashable,
        coroutine: abc.Coroutine
    ) -> None:
        """Await ``coroutine`` after ``delay`` seconds."""
        try:
            self._log.trace(f"Waiting {delay} seconds before awaiting coroutine for #{task_id}.")
            await asyncio.sleep(delay)

            # Use asyncio.shield to prevent the coroutine from cancelling itself.
            self._log.trace(f"Done waiting for #{task_id}; now awaiting the coroutine.")
            await asyncio.shield(coroutine)
        finally:
            # Close it to prevent unawaited coroutine warnings,
            # which would happen if the task was cancelled during the sleep.
            # Only close it if it's not been awaited yet. This check is important because the
            # coroutine may cancel this task, which would also trigger the finally block.
            state = inspect.getcoroutinestate(coroutine)
            if state == "CORO_CREATED":
                self._log.debug(f"Explicitly closing the coroutine for #{task_id}.")
                coroutine.close()
            else:
                self._log.debug(f"Finally block reached for #{task_id}; {state=}")

    def _task_done_callback(self, task_id: abc.Hashable, done_task: asyncio.Task) -> None:
        """
        Delete the task and raise its exception if one exists.

        If ``done_task`` and the task associated with ``task_id`` are different, then the latter
        will not be deleted. In this case, a new task was likely rescheduled with the same ID.
        """
        self._log.trace(f"Performing done callback for task #{task_id} {id(done_task)}.")

        scheduled_task = self._scheduled_tasks.get(task_id)

        if scheduled_task and done_task is scheduled_task:
            # A task for the ID exists and is the same as the done task.
            # Since this is the done callback, the task is already done so no need to cancel it.
            self._log.trace(f"Deleting task #{task_id} {id(done_task)}.")
            del self._scheduled_tasks[task_id]
        elif scheduled_task:
            # A new task was likely rescheduled with the same ID.
            self._log.debug(
                f"The scheduled task #{task_id} {id(scheduled_task)} "
                f"and the done task {id(done_task)} differ."
            )
        elif not done_task.cancelled():
            self._log.warning(
                f"Task #{task_id} not found while handling task {id(done_task)}! "
                f"A task somehow got unscheduled improperly (i.e. deleted but not cancelled)."
            )

        with contextlib.suppress(asyncio.CancelledError):
            exception = done_task.exception()
            # Log the exception if one exists.
            if exception:
                self._log.error(f"Error in task #{task_id} {id(done_task)}!", exc_info=exception)


TASK_RETURN = typing.TypeVar("TASK_RETURN")


def create_task(
    coro: abc.Coroutine[typing.Any, typing.Any, TASK_RETURN],
    *,
    suppressed_exceptions: tuple[type[Exception], ...] = (),
    event_loop: asyncio.AbstractEventLoop | None = None,
    **kwargs,
) -> asyncio.Task[TASK_RETURN]:
    """
    Wrapper for creating an :obj:`asyncio.Task` which logs exceptions raised in the task.

    If the ``event_loop`` kwarg is provided, the task is created from that event loop,
    otherwise the running loop is used.

    Args:
        coro: The function to call.
        suppressed_exceptions: Exceptions to be handled by the task.
        event_loop (:obj:`asyncio.AbstractEventLoop`): The loop to create the task from.
        kwargs: Passed to :py:func:`asyncio.create_task`.

    Returns:
        asyncio.Task: The wrapped task.
    """
    if event_loop is not None:
        task = event_loop.create_task(_coro_wrapper(coro), **kwargs)
    else:
        task = asyncio.create_task(_coro_wrapper(coro), **kwargs)

    _background_tasks.add(task)
    task.add_done_callback(_background_tasks.discard)
    task.add_done_callback(partial(_log_task_exception, suppressed_exceptions=suppressed_exceptions))
    return task


async def _coro_wrapper(coro: abc.Coroutine[typing.Any, typing.Any, TASK_RETURN]) -> None:
    """Wraps `coro` in a try/except block that will handle 90001 Forbidden errors."""
    try:
        await coro
    except Forbidden as e:
        await handle_forbidden_from_block(e)


def _log_task_exception(task: asyncio.Task, *, suppressed_exceptions: tuple[type[Exception], ...]) -> None:
    """Retrieve and log the exception raised in ``task``, if one exists and it's not suppressed."""
    with contextlib.suppress(asyncio.CancelledError):
        exception = task.exception()
        # Log the exception if one exists and it's not suppressed/handled.
        if exception and not isinstance(exception, suppressed_exceptions):
            log = logging.get_logger(__name__)
            log.error(f"Error in task {task.get_name()} {id(task)}!", exc_info=exception)
